package cn.estsh.i3plus.pojo.base.codemaker;

/**
 * @Description : 使用SnowFlake，生成分布式ID
 * @Reference : SnowFlake可以保证：
 * 所有生成的id按时间趋势递增
 * 整个分布式系统内不会产生重复id（因为有datacenterId和workerId来做区分）
 * @Author : alwaysfrin
 * @CreateDate : 2018-09-11 16:35
 * @Modify:
 **/
public class SnowflakeIdMaker {

    private long workerId;
    private long datacenterId;
    private long sequence = 0L;

    public SnowflakeIdMaker(){
        this.workerId = 1;
        this.datacenterId = 1;
        this.sequence = 0L;
    }

    public SnowflakeIdMaker(long workerId, long datacenterId){
        if (workerId > maxWorkerId || workerId < 0) {
            throw new IllegalArgumentException(String.format("worker Id can't be greater than %d or less than 0",maxWorkerId));
        }
        if (datacenterId > maxDatacenterId || datacenterId < 0) {
            throw new IllegalArgumentException(String.format("datacenter Id can't be greater than %d or less than 0",maxDatacenterId));
        }

        this.workerId = workerId;
        this.datacenterId = datacenterId;
    }

    private long twepoch = 1288834974657L;

    private long workerIdBits = 5L;
    private long datacenterIdBits = 5L;
    private long maxWorkerId = -1L ^ (-1L << workerIdBits); //30
    private long maxDatacenterId = -1L ^ (-1L << datacenterIdBits);  //30
    private long sequenceBits = 12L;

    private long workerIdShift = sequenceBits;
    private long datacenterIdShift = sequenceBits + workerIdBits;
    private long timestampLeftShift = sequenceBits + workerIdBits + datacenterIdBits;
    private long sequenceMask = -1L ^ (-1L << sequenceBits);

    private long lastTimestamp = -1L;

    private long getWorkerId(){
        return workerId;
    }

    private long getDatacenterId(){
        return datacenterId;
    }

    public synchronized long nextId() {
        long timestamp = timeGen();

        if (timestamp < lastTimestamp) {
            System.err.printf("clock is moving backwards.  Rejecting requests until %d.", lastTimestamp);
            throw new RuntimeException(String.format("Clock moved backwards.  Refusing to generate id for %d milliseconds",
                    lastTimestamp - timestamp));
        }

        if (lastTimestamp == timestamp) {
            sequence = (sequence + 1) & sequenceMask;
            if (sequence == 0) {
                timestamp = tilNextMillis(lastTimestamp);
            }
        } else {
            sequence = 0;
        }

        lastTimestamp = timestamp;
        return ((timestamp - twepoch) << timestampLeftShift) |
                (datacenterId << datacenterIdShift) |
                (workerId << workerIdShift) |
                sequence;
    }

    private long tilNextMillis(long lastTimestamp) {
        long timestamp = timeGen();
        while (timestamp <= lastTimestamp) {
            timestamp = timeGen();
        }
        return timestamp;
    }

    private long timeGen(){
        return System.currentTimeMillis();
    }

    //---------------测试---------------
    public static void main(String[] args) {
        SnowflakeIdMaker worker = new SnowflakeIdMaker();
        Runnable testRun = new Runnable() {
            @Override
            public void run() {
                for (int i = 0; i < 10; i++) {
                    System.out.println(worker.nextId());
                }
            }
        };
        Thread thread1 = new Thread(testRun);
        thread1.start();
        Thread thread2 = new Thread(testRun);
        thread2.start();
        Thread thread3 = new Thread(testRun);
        thread3.start();
        Thread thread4 = new Thread(testRun);
        thread4.start();
        Thread thread5 = new Thread(testRun);
        thread5.start();

        for (int i = 0; i < 10; i++) {
            System.out.println(worker.nextId());
        }
    }
}
